#include <string>
#include <assert.h>
#include <memory>
#include <utility>
#include <thread>
#include <mutex>
#include <iostream>
#include <stdio.h>
#include <jss/actor.hpp>

using jss::actor;
using jss::actor_ref;

std::mutex iom;

struct withdraw {
  std::string account;
  unsigned amount;
  actor_ref atm;
};

struct withdraw_ok {};

struct withdraw_denied {};

struct cancel_withdrawal {
  std::string account;
  unsigned amount;
};

struct withdrawal_processed {
  std::string account;
  unsigned amount;
};

struct card_inserted {
  std::string account;
};

struct digit_pressed {
  char digit;
};

struct clear_last_pressed {};

struct eject_card {};

struct withdraw_pressed {
  unsigned amount;
};

struct cancel_pressed {};

struct issue_money {
  unsigned amount;
};

struct verify_pin {
  std::string account;
  std::string pin;
  actor_ref atm;
};

struct pin_verified {};

struct pin_incorrect {};

struct display_enter_pin {};

struct display_enter_card {};

struct display_insufficient_funds {};

struct display_withdrawal_cancelled {};

struct display_pin_incorrect_message {};

struct display_withdrawal_options {};

struct get_balance {
  std::string account;
  actor_ref atm;
};

struct balance {
  unsigned amount;
};

struct display_balance {
  unsigned amount;
};

struct balance_pressed {};

class atm {
  actor_ref bank;
  actor_ref interface_hardware;
  void (atm::*state)();

  std::string account;
  unsigned withdrawal_amount;
  std::string pin;

  void process_withdrawal() {
    actor::receive()
      .match<withdraw_ok>(
        [&](withdraw_ok const& ) {
          interface_hardware<<issue_money{withdrawal_amount};
          bank<<withdrawal_processed{
            account,withdrawal_amount};
          state=&atm::done_processing;
        })
      .match<withdraw_denied>(
        [&](withdraw_denied const& ) {
          interface_hardware<<display_insufficient_funds();
          state=&atm::done_processing;
        })
      .match<cancel_pressed>(
        [&](cancel_pressed const& ) {
          bank<<cancel_withdrawal{account,withdrawal_amount};
          interface_hardware<<display_withdrawal_cancelled();
          state=&atm::done_processing;
        });
  }

  void process_balance() {
    actor::receive()
      .match<balance>(
        [&](balance const& msg) {
          interface_hardware<<display_balance{msg.amount};
          state=&atm::wait_for_action;
        })
      .match<cancel_pressed>(
        [&](cancel_pressed const& ) {
          state=&atm::done_processing;
        });
  }

  void wait_for_action() {
    interface_hardware<<display_withdrawal_options();
    actor::receive()
      .match<withdraw_pressed>(
        [&](withdraw_pressed const& msg) {
          withdrawal_amount=msg.amount;
          bank<<withdraw{account,msg.amount,actor::self()};
          state=&atm::process_withdrawal;
        })
      .match<balance_pressed>(
        [&](balance_pressed const& ) {
          bank<<get_balance{account,actor::self()};
          state=&atm::process_balance;
        })
      .match<cancel_pressed>(
        [&](cancel_pressed const& ) {
          state=&atm::done_processing;
        });
  }

  void verifying_pin() {
    actor::receive()
      .match<pin_verified>(
        [&](pin_verified const& ) {
          state=&atm::wait_for_action;
        })
      .match<pin_incorrect>(
        [&](pin_incorrect const& ) {
          interface_hardware<<display_pin_incorrect_message();
          state=&atm::done_processing;
        })
      .match<cancel_pressed>(
        [&](cancel_pressed const&) {
          state=&atm::done_processing;
        });
  }


  void getting_pin() {
    actor::receive()
      .match<digit_pressed>(
        [&](digit_pressed const& msg) {
          unsigned const pin_length=4;
          pin+=msg.digit;
          if(pin.length()==pin_length) {
            bank<<verify_pin{account,pin,actor::self()};
            state=&atm::verifying_pin;
          }
        })
      .match<clear_last_pressed>(
        [&](clear_last_pressed const&) {
          if(!pin.empty()) {
            pin.resize(pin.length()-1);
          }
        })
      .match<cancel_pressed>(
        [&](cancel_pressed const&) {
          state=&atm::done_processing;
        });
  }

  void waiting_for_card() {
    interface_hardware<<display_enter_card();
    actor::receive()
      .match<card_inserted>(
        [&](card_inserted const& msg) {
          account=msg.account;
          pin="";
          interface_hardware<<display_enter_pin();
          state=&atm::getting_pin;
        });
  }

  void done_processing() {
    interface_hardware<<eject_card();
    state=&atm::waiting_for_card;
  }
public:
  atm(actor_ref bank_,
      actor_ref interface_hardware_):
    bank(bank_),interface_hardware(interface_hardware_) {}

  void operator()() {
    state=&atm::waiting_for_card;
    for(;;) {
      (this->*state)();
    }
  }
};

class bank_machine {
  unsigned balance;
public:
  bank_machine():
    balance(199) {}

  void operator()() {
    for(;;) {
      actor::receive()
        .match<verify_pin>(
          [&](verify_pin const& msg) {
            if(msg.pin=="1937") {
              msg.atm<<pin_verified();
            }
            else {
              msg.atm<<pin_incorrect();
            }
          })
        .match<withdraw>(
          [&](withdraw const& msg) {
            if(balance>=msg.amount) {
              msg.atm<<withdraw_ok();
              balance-=msg.amount;
            }
            else {
              msg.atm<<withdraw_denied();
            }
          })
        .match<get_balance>(
          [&](get_balance const& msg) {
            msg.atm<<::balance{balance};
          })
        .match<withdrawal_processed>(
          [&](withdrawal_processed const&) {
          })
        .match<cancel_withdrawal>(
          [&](cancel_withdrawal const&) {
          });
    }
  }
};

class interface_machine {
public:
  void operator()() {
    for(;;) {
      actor::receive()
        .match<issue_money>(
          [&](issue_money const& msg) {
            std::lock_guard<std::mutex> lk(iom);
            std::cout<<"Issuing "<<msg.amount<<std::endl;
          })
        .match<display_insufficient_funds>(
          [&](display_insufficient_funds const&) {
            std::lock_guard<std::mutex> lk(iom);
            std::cout<<"Insufficient funds"<<std::endl;
          })
        .match<display_enter_pin>(
          [&](display_enter_pin const&) {
            std::lock_guard<std::mutex> lk(iom);
            std::cout<<"Please enter your PIN (0-9)"
                     <<std::endl;
          })
        .match<display_enter_card>(
          [&](display_enter_card const&) {
            std::lock_guard<std::mutex> lk(iom);
            std::cout<<"Please enter your card (i)"
                     <<std::endl;
          })
        .match<display_balance>(
          [&](display_balance const& msg) {
            std::lock_guard<std::mutex> lk(iom);
            std::cout<<"The balance of your account is "
                     <<msg.amount<<std::endl;
          })
        .match<display_withdrawal_options>(
          [&](display_withdrawal_options const&) {
            std::lock_guard<std::mutex> lk(iom);
            std::cout<<"Withdraw 50? (w)"<<std::endl;
            std::cout<<"Display Balance? (b)"<<std::endl;
            std::cout<<"Cancel? (c)"<<std::endl;
          })
        .match<display_withdrawal_cancelled>(
          [&](display_withdrawal_cancelled const&) {
            std::lock_guard<std::mutex> lk(iom);
            std::cout<<"Withdrawal cancelled"<<std::endl;
          })
        .match<display_pin_incorrect_message>(
          [&](display_pin_incorrect_message const&) {
            std::lock_guard<std::mutex> lk(iom);
            std::cout<<"PIN incorrect"<<std::endl;
          })
        .match<eject_card>(
          [&](eject_card const&) {
            std::lock_guard<std::mutex> lk(iom);
            std::cout<<"Ejecting card"<<std::endl;
          });

    }
  }
};


int main() {
  actor bank{bank_machine()};
  actor interface_hardware{interface_machine()};

  actor machine{atm(bank,interface_hardware)};

  bool quit_pressed=false;

  while(!quit_pressed) {
    char c=getchar();
    switch(c) {
    case '0':
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
      machine<<digit_pressed{c};
      break;
    case 'b':
      machine<<balance_pressed();
      break;
    case 'w':
      machine<<withdraw_pressed{50};
      break;
    case 'c':
      machine<<cancel_pressed();
      break;
    case 'q':
      quit_pressed=true;
      break;
    case 'i':
      machine<<card_inserted{"acc1234"};
      break;
    }
  }

  bank.stop();
  machine.stop();
  interface_hardware.stop();
}

